import logging
import time
from logstash_async.handler import AsynchronousLogstashHandler
from logstash_async.formatter import LogstashFormatter
from egune.interfaces import ActorMessage, ApiRequest, Command, Interface, UserMessage
from typing import Callable, Dict, Any, List
import traceback
import json


logger = logging.getLogger('')


def init_logger(config):
    global logger
    logger = logging.getLogger(config["name"])
    logger.setLevel(logging.INFO)
    logstash_formatter = LogstashFormatter(message_type=f"egune-{config['name']}")
    logstash_handler = AsynchronousLogstashHandler(config["host"], config["port"], None)
    logstash_handler.setFormatter(logstash_formatter)
    logger.addHandler(logstash_handler)
    logger.info(f"Started {config['name']}", extra={
        "log_event": "system started"
    })


def log_request(request_name, input_prep=None, output_prep=None):
    def ultra_wrapped(func):
        def wrapped(data):
            s = time.time()
            original_input = data
            try:
                if input_prep is not None:
                    data = input_prep(data)
                if isinstance(data, tuple):
                    result = func(*data)
                else:
                    result = func(data)
                if output_prep is not None:
                    result = output_prep(result)
                logger.info("Processed", extra={
                    "log_event": request_name,
                    "process_input": original_input,
                    "process_output": result,
                    "time": time.time() - s
                })
                return result
            except Exception as e:
                logger.error("Failed", extra={
                    "log_event": request_name + ":Failed",
                    "process_input": original_input,
                    "process_output": traceback.format_exc(),
                    "time": time.time() - s
                })
                return e
        return wrapped
    return ultra_wrapped


def log(extra):
    logger.info("custom", extra=extra)



from egune.requests import (  # noqa: 401
    ButtonQuestion,
    YesNoQuestion,
    OpenQuestion,
    Fail,
    Success,
    Tell,
    TellCustom,
    TellPreparedMessage,
    Do,
    MultiSelectQuestion,
    CheckboxQuestion,
    Form,
    FormQuestionTypes,
    ActorMessage,
    ActionServerRequest
)
from datetime import datetime
from cachetools import TTLCache
import requests

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class Egune:
    def __init__(self, egune_host, egune_port, api_token, is_test=False):
        self.egune_home = f"http://{egune_host}:{egune_port}/"
        self.api_token = api_token
        self.is_test = is_test
        self.cache = TTLCache(9999, 60)

    def tell_user_now(self, user_id: str, message: ActorMessage):
        request = TellCustom()
        request.variables = {"msg": message.to_dict()}
        self.send_request_now(user_id, request)

    def send_request_now(self, user_id: str, request: ActionServerRequest):
        request_dict = request.publish()
        request_dict["app_id"] = self.api_token
        request_dict["user_id"] = user_id
        request_dict["user_message"] = {
            "user_id": user_id
        }
        request_dict["operation"] = "process"
        if self.is_test:
            print("EGUNE SENT MESSAGE:", request_dict)
        else:
            requests.post(
                self.egune_home + "app",
                json=request_dict
            )

    def create_ontime_event(self, user_id, timestamp: datetime, name="on time single event", data={}):
        event = {
            "type": "on-time",
            "user_id": user_id,
            "app_id": self.api_token,
            "time": timestamp.strftime("%Y-%m-%d %H:%M:%S"),
            "name": name,
            "data": data,
            "operation": "add_event"
        }
        if self.is_test:
            print("EGUNE EVENT:", event)
        else:
            requests.post(
                self.egune_home + "app",
                json=event
            )

    def create_weekly_event(self, user_id, weekday, hour, minute, name="weekly repeating event", data={}):
        event = {
            "type": "weekly",
            "user_id": user_id,
            "app_id": self.api_token,
            "hour": hour,
            "minute": minute,
            "name": name,
            "weekday": weekday,
            "data": data,
            "operation": "add_event"
        }
        if self.is_test:
            print("EGUNE EVENT:", event)
        else:
            requests.post(
                self.egune_home + "app",
                json=event
            )

    def create_monthly_event(self, user_id, day, hour, minute, name="monthly repeating event", data={}):
        event = {
            "type": "monthly",
            "user_id": user_id,
            "app_id": self.api_token,
            "hour": hour,
            "minute": minute,
            "name": name,
            "day": day,
            "data": data,
            "operation": "add_event"
        }
        if self.is_test:
            print("EGUNE EVENT:", event)
        else:
            requests.post(
                self.egune_home + "app",
                json=event
            )

    def save_state(self, user_id, vals):
        self.cache[user_id] = vals

    def get_state(self, user_id):
        return self.cache.get(user_id, None)

    def add_data(self, train, test):
        request = {
            "operation": "add_data",
            "train": train,
            "test": test,
            "app_id": self.api_token
        }
        requests.post(self.egune_home + "app", json=request)
